 //
//Date: 2010/11/14;
//Author: Peter Hansen, Zhuo Huang and Howard Shek
//Purpose: Codes to define an the log-linear RealGARCH(1,2) class
//


#include <oxstd.h>
#include <oxfloat.h>
#import <database>
#import <maximize>
#include <oxprob.h> 
#import  <maxsqp>  
  
   

class RealGARCH: Database   
{

	decl m_PI; //Define Constant PI;

	
   //Define member variables about parameters 
   decl m_vlo ;
   decl m_vhi;

   
	decl m_vPar;	//Parameters:
	decl m_Est;		//Estimates
	decl m_SE;		//Standard deviation
	decl m_tValue;	//t-value
	

  // Define member variables about data;

   decl m_mData ;
   decl m_mY;	   	// dependent variable m_cT x 1
   decl m_mX;	   	// exogenous explanatory variables
   decl m_vDate;	// date information

   decl m_T0;		//number of observations of the input data
   decl m_T;		//number of observations of the constructed data for estiamtion
   decl m_TValid;	// number of in-sample observations
  
   decl m_Pai;
   
   //Define output;
   	decl m_iResult; // result from MaxBFGS

  
     // Define member variables about likelihoods;
	
	decl m_dLogLik,m_dLogLik1;	// log-likelihood vector	
	decl m_logL,m_logL1;
	decl m_logL_os,m_logL1_os;

	//Define member variables about vectors in implied by the model
	decl m_h;	   //vector of conditional variance
	decl m_v;	   //vector of u 
	decl m_u;	   //vector of v
	decl m_z;	   //vector of z
	decl m_tau_z;  //vector of tau(z)
	
 
  	 
	// Constructor
	RealGARCH( const Data);

	//Define member functions
    InitPar();
 	DoEstimate();	
 	Output( );
	 
	logLikFuncInternal(const vPFull, const adFunc);
	logLikFunc(const vP, const adFunc, const avScore, const amHess);

	//constaint fucntion for estimation
 	cfunc_gt0(const avF,const vP);

	//function to calculate robust standard error
	RobustSE(const vP );
	 
} 

 
RealGARCH::RealGARCH( const Data ) {
 
	Database();		// initialize base class
	
	decl m_cLag=3;	//use the first 3 obs to construct lag variables
	
	m_mData=Data;
	m_T0=rows(m_mData);	//number of obs of the raw data
	
 
	m_vDate=	m_mData[m_cLag:m_T0-1][0];
	
	m_mY=m_mData[m_cLag:m_T0-1][1];		  //return

	//x with 3 lags;
	m_mX=m_mData[m_cLag:m_T0-1][2]~m_mData[m_cLag-1:m_T0-2][2]~m_mData[m_cLag-2:m_T0-3][2]~m_mData[m_cLag-3:m_T0-4][2];

	m_mData= m_vDate~m_mY~m_mX;


	m_T=rows(m_mData);	//number of obs of the constructed data

//	println(m_mData[0:10][]);
 
    }
 


//Initialize parameters;

RealGARCH::InitPar( ) {    
  
	decl h0= -0.28921;

	decl omega= 0.041246;
	
	decl beta= 0.70121;
	
	decl gamma=  0.45068  | -0.17604;

	decl xi= -0.17999 ;
	
	decl phi= 1.0375 ;
	
	decl sigU=   0.14537 ;

	decl tau=-0.067810 |   0.070158 ;
	

	//Defne bounds of parameters;
 	decl vlo_h0=<-10>  ;
	decl vhi_h0=<3>   ;
 	
 	 decl vlo_omega=-3    ; 
	 decl vhi_omega=3   ;
	 
	 decl vlo_beta=-1    ;
	 
	 decl vhi_beta=2   ;
	
	 decl vlo_gamma=<-1> * ones(2,1) ;
	 decl vhi_gamma=< 2 >  * ones(2,1) ;

	 decl vlo_xi=-10   ;
	 decl vhi_xi= 3   ;

	 decl vlo_phi=0.5 ;
	 decl vhi_phi=5   ;

	 decl vlo_tau=< -0.5 > * ones(2,1) ;
     decl vhi_tau=< 0.5 > * ones(2,1)  ;

	 decl vlo_sigU=0.01  ;
	 decl vhi_sigU=4   ;
	
 
    decl   vP=h0|omega|beta|gamma|xi|phi|tau|sigU;
	
	decl  vlo= vlo_h0|vlo_omega|vlo_beta|vlo_gamma|vlo_xi|vlo_phi|vlo_tau|vlo_sigU;
    decl  vhi= vhi_h0|vhi_omega|vhi_beta|vhi_gamma|vhi_xi|vhi_phi|vhi_tau|vhi_sigU;
	
 	 
	
	 m_vlo	=vlo ;
	 m_vhi	= vhi  ;
	
	 println("vP", vlo~vP~vhi);
 
	 return vP;
	
}	

//logLikelihood function 
RealGARCH::logLikFunc(const vP, const adFunc, const avScore, const amHess){


	decl adValue;
	
	logLikFuncInternal( vP, &adValue);
	
	adFunc[0]=adValue;
	
	return TRUE;

}


RealGARCH::logLikFuncInternal(const vP0, const adFunc ) {

 
 	decl p=1;	  //lag of h
	
	decl q=0;	  //lag of ret^2

 
	decl r=2;		   //lag of x
	
	decl vP=vP0; 
 
	decl tau=0,res_z=0,res_u=0,res_v,tau_z=0;
	
	decl y = m_mY[][0];		 //return
	 
	decl logXs= log(m_mX[][0:r-1]) ;   //log(x) and its lags
	 
		
	decl h_ini = exp(vP[0]);   //h0, treated as a free pamameter

		
	decl beta = vP[2];

	decl omega=vP[1];
	
 	decl gamma=vP[3:4];
	
    decl lamda= omega|beta|gamma;
 	
	decl  xi=vP[5];
	decl  phi=vP[6];
 
       tau= vP[7:8];	 

	  decl   sigU= sqrt(vP[9]) ;	 


   //	Initialize variables;  
	decl logL = 0, logL1=0, T = sizeof(y);
	  
	decl z = zeros(4,1);	 //z=(1,log(h_t-1),log(X_t-1),loh(x_t-2))
	
	decl h = zeros(T,1);
	
	decl  logLs=zeros(T,1);
	decl  logL1s=zeros(T,1);

	
	decl  vector_v=zeros(T,1);
	decl  vector_u=zeros(T,1);
	decl  vector_tau_z=zeros(T,1);


	
	h[0]= h_ini;	 
  
	decl t, h_t;
	
	z[0] = 1;	//z=(1,log(h_t-1),log(X_t-1),loh(x_t-2))
	 	
		// For the first period;	  		
		h_t = h[0];		
		res_z= y[0]/sqrt(h_t);		
  		tau_z= tau[0]*res_z + tau[1]* (res_z^2-1) ;	 
		res_u=   logXs[0][0] -xi-phi* log(h_t)- tau_z   ;
		res_v= 	res_u   + 	tau_z   ; 

		  	
 	 		vector_v[0]=	res_v;
			vector_u[0]=   res_u;
			vector_tau_z[0]=  	tau_z;
		   		   
		   logLs[0] =     (-1/2)*log(h_t) +  log(densn(res_z))+  log( densn(res_u/ sigU )) -  log(sigU) ;
           logL1s[0] =    (-1/2)*log(h_t) +  log(densn(res_z))  ;
		  
	 	   z[1] =log(h_t); 	
	 
		  	  
 
 	//From the 2nd period
	
	for (t=1; t<T; ++t){
	     
      z[2:3] = (logXs[t-1][0:r-1])'; 
	
      h_t = exp(  lamda'*z) ; 
		
	   h[t]=h_t;
	  
	   	res_z= y[t]/sqrt(h_t);
		
		tau_z= tau[0]*res_z + tau[1]* (res_z^2-1) ;		
		 

		res_u=   logXs[t][0] -xi-phi* log(h_t)- tau_z   ;
		res_v= 	res_u   + 	tau_z   ; 
						 
 		 	
 	 		vector_v[t]=	 res_v;
			vector_u[t]=     res_u;
			vector_tau_z[t]= tau_z;
	 
		
		  logLs[t] =     (-1/2)*log(h_t) +  log(densn(res_z)) +    log(densn(res_u/ sigU )) -  log(sigU)  ;
		 
          logL1s[t] =    (-1/2)*log(h_t) +  log(densn(res_z))   ;
	    	  			
 
	       z[1] =log(h_t); 	
		
	}
	
    	m_dLogLik=  logLs;
		m_dLogLik1=  logL1s;

 
		m_h=h;
		m_z=y./sqrt(h); //standardized return
		m_v=vector_v;
		m_u= vector_u/sigU;		
 		m_tau_z=vector_tau_z; 
		m_T=T;
		m_Pai= sumc(beta)+phi*sumc(gamma);
			
  	 decl logLsValid=logLs[0:(T-1)];
	 decl logL1sValid=logL1s[0:(T-1)];
   
	  //define in-sample and out-of-sample
		m_logL_os =    sumc( selectifr(logLsValid,m_vDate.> 20080000)  );
	    m_logL1_os =    sumc( selectifr(logL1sValid, m_vDate.> 20080000 )  );

	    logLsValid= selectifr(logLsValid,m_vDate.< 20080000 )  	;
		logL1sValid= selectifr(logL1sValid, m_vDate.< 20080000 )  	;
	
	m_TValid=rows(logLsValid);
	
	logL=   sumc(logLsValid);
	logL1=   sumc(logL1sValid);
	
	
	m_logL=logL;
	m_logL1=logL1;

	
	 adFunc[0] =	logL;
	
	 return 1;
}




//Constraint on the persistent parameter
RealGARCH::cfunc_gt0(const avF,const vP0)
   {

   	   decl vP=vP0;
 
	   decl  beta=0,gamma=0,phi=1; 

        beta = vP[2];  
	 
	   gamma = vP[3:4]; 

	   phi = vP[5]; 
	  
	  avF[0]=matrix( 0.9999-  sumc(beta)-phi*sumc(gamma) );  // . 

	 return TRUE;

   }

 RealGARCH::DoEstimate(){

      	decl vP=InitPar();
 		
  	    decl dfunc,Hessian;
 		
 	    logLikFunc(vP, &dfunc,0,0);
 			
    	println("the initial log likelihood is:",dfunc );


 	    MaxControl(300, 40);

      	MaxControlEps( 10e-10, 10e-6);

        decl result= MaxSQP(logLikFunc, &vP, &dfunc, 0, 1,cfunc_gt0, 0 ,m_vlo,m_vhi);
 
	    println(MaxConvergenceMsg( result));

	    logLikFunc(vP, &dfunc,0,0);
 		
		println("Optimal logL:",dfunc);
		
		println("Optimal logL1:",m_logL1);
	    println("Estimates:",vP');

			
		//numerical score and Hessian;
		decl avScore  ;
		Num1Derivative( logLikFunc, vP,  &avScore);

			println("Numerical Score:",avScore');
			
		Num2Derivative(logLikFunc, vP, &Hessian) ;

			logLikFunc(vP, &dfunc,0,0);
 			m_Est=vP';

			
  			//SE using inverse Hessian;
		    decl stdErr =  sqrt( diagonal(invertgen(-1*Hessian)));	
			println("SE:", stdErr );
	     	print("The t-values  I-1 numerical are: ", vP'./ stdErr );
		

			//Robust SE
			
 			  decl robustSE=  RobustSE(vP);

			  robustSE[0]= stdErr[0];//For SE for logh0, I still use that from inverse Hessian

			    println("Analytical SIGMA_I-1JI-1",   robustSE);
	  
			   //Output robust SE and t value
			   m_SE=robustSE;
			   m_tValue=  vP'./ robustSE ;
			    println("t-Value using Analytical SIGMA_I-1JI-1",   m_tValue);
	  
			   
						
			   
		return 1;
		

 }
 
 RealGARCH::Output(){
	    //Output estimates		 
		decl OutFileEst="Results\Est.xls";
		decl EstNames={"logh0","Omega","Beta","Gamma1","Gamma2","Xi","Phi","Tau1","Tau2","SigU^2"};
		savemat(OutFileEst,m_Est|m_SE|m_tValue,EstNames);
 
	  //Output summary statistics
      decl OutFileSum= "Results\Summary.xls";
      decl SumNames={"T","T Valid","Pai","logL","logL1","logL_OS","logL1_OS"};
      savemat(OutFileSum, m_T~m_TValid~m_Pai~m_logL~m_logL1~m_logL_os~m_logL1_os,SumNames);

	  //Output variables;
		decl OutVars= "Results\Variables.xls";
	 	decl VarNames={"Date","logLs","logL1s","h","z", "v","u","tau_z","Ret","RK"};
        savemat(OutVars,m_vDate~m_dLogLik~m_dLogLik1~m_h~m_z~ m_v~m_u~m_tau_z~m_mY~m_mX[][0],VarNames );

  
      return TRUE;


 }
 

 // The funcion to return the J matrix that will be used to calculate robust SE.
    RealGARCH::RobustSE(const vP ){

	decl temp;

	decl omega = vP[1];
		
	decl beta = vP[2];

	decl lamda, gamma;
	
 
      gamma=vP[3:4];
	
	  lamda=omega|beta|gamma;
 	
	
	   decl  xi,phi, sigU,tau;
	 
  
       xi=vP[5];
	  
	   
	   phi=vP[6];

	   tau= vP[7:8];	 

	   sigU= sqrt(vP[9]);
	 
	
  

	   decl k= rows(vP)-1;	  //we don't need log(h0) now;
	   
	   decl T_Valid=m_TValid;
		
 
 
	   
	   decl I=zeros(k,k)  , J=zeros(k,k), Jt, It;

	   decl  iT, u,tau_z,z,h,x,m,g, G, psi,  i,j ,u_d,u_dd	;
	   decl	 It11,It12,It13,It21,It22,It23,It31,It32,It33;
	   decl Scores=zeros(m_TValid,k);
 
	   decl gh=zeros(1,1);
	   decl gx=zeros(2,1);
	   decl h_d,  h_d1 ,h_d2  ;
	   decl h_dd, h_dd1,h_dd2; 


 	  decl g0, G0;
 
	  
	  	    g0= (1~ meanc(log(m_h))~  ones(1,2)* meanc(log(m_mX[][0]))     )'	 ;
			
			h_d= 0* g0 / (1- beta ) ;
    			
			h_d1=h_d;
			h_d2=h_d;

	 	
			G0=  (zeros(4,1) ~  h_d1 ~   zeros(4,2) ) +     (zeros(4,1)  ~  h_d1 ~   zeros(4,2) )'   ; 

			h_dd= 0*  G0/ (1- beta[0]);
			h_dd1=h_dd;
			h_dd2=h_dd;
	 
	 
 
 		
	  for (iT=0;iT<m_TValid;iT++) {

	   
	   for(i=0;i<1;i++){

	   		 if (iT<2) gh[i]= log(m_h[iT]);
	   
	   		 else gh[i]=log(m_h[iT-i-1]);
	   
	   }

	   for(i=0;i<2;i++){
	   		 if (iT<2) gx[i]=log(m_mX[iT][0]);
	    	 else gx[i]=log(m_mX[iT-i-1][0]);

	   }

	   g=(1~gh'~gx')';

	   
	    h_d = g + beta' * h_d1;
	   
	 
	    G= (zeros(4,1) ~  h_d1 ~   zeros(4,2) ) +     (zeros(4,1)  ~  h_d1 ~   zeros(4,2) )'   ; 
 
        h_dd = G + beta' * h_dd1;
 
 
	
	   
	   u=m_u[iT]*sigU;
	   
	   z=m_z[iT];

	   m=(1~log(m_h[iT])~z~(z^2-1))';

	   u_d=-phi+0.5*tau[0]*z+tau[1]*z^2;
	   
	   u_dd=-0.25*tau[0]*z-tau[1]*z^2;


	   temp= (1/2) *(	 (1-z^2+ 2*u*u_d/(sigU^2) )* h_d |-2* (u/(sigU^2))*m |(sigU^2-u^2)/sigU^4    );

	   Scores[iT][] =   temp'  ;
	   
       Jt=  temp*temp';

	    decl vector_b= (0~1~ -0.5*z~ -z^2)';
	   
	   
	     It11= -0.5* (  z^2+ 2*( u_d^2+u*u_dd )/(sigU^2)   ) *h_d*h_d' -0.5*( 1-z^2+2*u*u_d/(sigU^2)   )*h_dd ;
	     It21= (1/sigU^2)* (u_d*m  + u* vector_b  )*h_d';
	     It22= -m*m'/(sigU^2);
	     It31= u*u_d  * h_d' /( sigU^4);

	     // It32= - u*m' /(sigU^4);								//The mean is zero
	   
	   	  It32= zeros(1,4);
		  
	     // It33= 0.5 * (2*u^2-sigU^2)/(sigU^6);		  
		 
		 It33=-0.5 /(sigU^4);
		  
	     It12=It21';
	     It23=It32';

	     It13=It31';


	    It=  ( It11~It12~It13) |(It21~It22~It23) |(It31~It32~It33);

		I=I+It;

		J=J+Jt	  ;

	   h_d1= h_d;
	   h_d2=h_d1;
	   
	   h_dd1= h_dd;
	   h_dd2=h_dd1;
	   
	   }


	   I=I/T_Valid;
	   J=J/T_Valid;
	   
 
 	   
	   decl SIGMA_J= sqrt( diagonal(invert(J))/T_Valid );

	    println("Analytical SIGMA_J-1", 1~ SIGMA_J);
		
 
	   decl SIGMA_I= sqrt( diagonal(invertgen(-I))/T_Valid );

	   
 	 
	   decl SIGMA_IJI= sqrt( diagonal(  invert(-I) * J* invert(-I)   )/T_Valid );	   
  
      //We actually use U from numerical Hessian;	 But the results are very close;
   	   // return J;

	   return (1~ SIGMA_IJI);
 }
